import re

import socket
from socket import gaierror as SocketError

from dns import resolver

import logging

logger = logging.getLogger("srv_hijacker")


def resolve_ip_for_target(rrsets, target):
    for rrset in rrsets:
        if rrset.name == target:
            return rrset.items[0].address

    raise SocketError("Couldn't find SRV record for target %s" % target)


def resolve_srv_record(old_host, resolver):
    ans = resolver.query(old_host, "SRV")

    new_port = ans[0].port
    new_host = resolve_ip_for_target(ans.response.additional, ans[0].target)

    logger.debug(
        "Resolved SRV record for host %s: (%s:%s)", old_host, new_host, new_port
    )

    return new_host, new_port


original_socket_getaddrinfo = socket.getaddrinfo


def patched_socket_getaddrinfo(host_regex, srv_resolver):
    """
    Returns a function that behaves like `socket.getaddrinfo`.

    host_regex:

    The regex to match a host against. If this regex matches the host
    we hit srv_resolver to fetch the new host + port
    """

    def patched_f(host, port, family=0, type=0, proto=0, flags=0):
        if re.search(host_regex, host):
            logger.debug("TCP host %s matched SRV regex, resolving", host)
            host, port = resolve_srv_record(host, srv_resolver)
        else:
            logger.debug("TCP host %s did not match SRV regex, ignoring", host)

        return original_socket_getaddrinfo(host, port, family, type, proto, flags)

    return patched_f


def hijack(host_regex, srv_dns_host=None, srv_dns_port=None):
    """
    Usage:

    ```
    srv_hijacker.hijack(
        host_regex=r'service.consul$',
        srv_dns_host='127.0.0.1',
        srv_dns_port=8600
    )
    ```
    """
    srv_resolver = resolver.Resolver()
    if srv_dns_host:
        srv_resolver.nameservers = [srv_dns_host]
    if srv_dns_port:
        srv_resolver.port = int(srv_dns_port)

    socket.getaddrinfo = patched_socket_getaddrinfo(host_regex, srv_resolver)



import pytest
import requests

import srv_hijacker

CONSUL_HOST = "127.0.0.1"
CONSUL_DNS_PORT = "8600"
CONSUL_API_PORT = "8500"
CONSUL_API_URL = f"{CONSUL_HOST}:{CONSUL_API_PORT}"


def test_hijack(consul_http_service_url):
    def test_http(url):
        response = requests.get(url)
        assert response.status_code == 200

    def test_http_direct():
        test_http("https://httpbin.org/get")

    def test_http_hijacked():
        test_http(consul_http_service_url)

    # Before 'hijack' is run, direct requests must pass
    test_http_direct()

    # Before 'hijack' is run, hijacked requests must fail
    with pytest.raises(requests.exceptions.ConnectionError):
        test_http_hijacked()

    srv_hijacker.hijack(
        host_regex=r"service.consul$",
        srv_dns_host=CONSUL_HOST,
        srv_dns_port=CONSUL_DNS_PORT,
    )

    # Now that the monkey patching is done, both should succeed
    test_http_direct()
    test_http_hijacked()

    # Patching again shouldn't cause issues
    srv_hijacker.hijack(
        host_regex=r"service.consul$",
        srv_dns_host=CONSUL_HOST,
        srv_dns_port=CONSUL_DNS_PORT,
    )

    test_http_direct()
    test_http_hijacked()


@pytest.fixture
def consul_http_service_url():
    # Let's register a service named 'test'  to point back to consul itself
    register_service_on_consul("test-http", CONSUL_HOST, CONSUL_API_PORT)

    # We can now use a consul endpoint itself for testing.
    return "http://test-http.service.consul/v1/agent/services"


def register_service_on_consul(service_name, service_host, service_port):
    url = f"http://{CONSUL_API_URL}/v1/agent/service/register"
    response = requests.put(
        url,
        json={"Name": service_name, "Address": service_host, "Port": int(service_port)},
    )

    assert response.status_code == 200



from setuptools import setup
from setuptools import find_packages

setup(
    name="srv_hijacker",
    version="0.0.6",
    description="Patch urllib3 to query a certain DNS server for SRV records",
    url="https://github.com/rohitpaulk/srv_hijacker",
    author="Paul Kuruvilla",
    author_email="paul.kuruvilla@shuttl.com",
    license="MIT",
    packages=find_packages(),
    classifiers=["Programming Language :: Python :: 3.7"],
    install_requires=["dnspython"],
)
